#include "pch.h"
#include "AudioSample.h"
#include "LoadedAudioFileInfo.h"
#include "AudioFilePlayer.h"

using namespace DeeJay;
using namespace Platform;
using namespace Windows::Storage::Streams;

AudioFilePlayer::AudioFilePlayer(IXAudio2* pXAudio2, LoadedAudioFileInfo loadedAudioFileInfo)
{
    // Create the source voice
    HRESULT hresult = pXAudio2->CreateSourceVoice(&pSourceVoice, &loadedAudioFileInfo.waveFormat,
                                                  XAUDIO2_VOICE_NOPITCH, 1.0f, this);
    if (FAILED(hresult))
        throw ref new COMException(hresult, "IXAudio2->CreateSourceVoice failure");

    // Initialize several fields
    pAudioBuffer = loadedAudioFileInfo.pBuffer;
    audioBufferLength = loadedAudioFileInfo.bufferLength;
    audioBufferIndex = 0;
    samplesPerSecond = loadedAudioFileInfo.waveFormat.nSamplesPerSec;
    speedRatio = 0;

    // Submit two alternating buffers, 1/100th second each
    playerBufferLength = samplesPerSecond / 100;
    playerBuffer1 = new AudioSample[playerBufferLength];
    playerBuffer2 = new AudioSample[playerBufferLength];

    SubmitBuffer(playerBuffer1, playerBufferLength);
    SubmitBuffer(playerBuffer2, playerBufferLength);

    // Start the voice playing
    pSourceVoice->Start();
}

AudioFilePlayer::~AudioFilePlayer()
{
    // Destroy this voice. We're all done here
    pSourceVoice->DestroyVoice();

    // Delete big audio buffer obtained from constructor args
    delete[] pAudioBuffer;

    // Delete small player buffers allocated in constructor
    delete[] playerBuffer1;
    delete[] playerBuffer2;
}

// Public methods
void AudioFilePlayer::SetSpeedRatio(double speedRatio)
{
    this->speedRatio = speedRatio;
}

void AudioFilePlayer::SetPositionFraction(double fraction)
{
    fraction = max(0, min(1, fraction));
    audioBufferIndex = fraction * audioBufferLength;
}

double AudioFilePlayer::GetSecondsElapsed()
{
    return audioBufferIndex / samplesPerSecond;
}

// Private methods
void AudioFilePlayer::SubmitBuffer(AudioSample * pPlayerBuffer, int playerBufferLength)
{
    for (int i = 0; i < playerBufferLength; i++)
    {
        pPlayerBuffer[i] = InterpolateSamples();
        audioBufferIndex = min(audioBufferLength, max(0, audioBufferIndex + speedRatio));
    }

    XAUDIO2_BUFFER buffer = {0};
    buffer.AudioBytes = playerBufferLength * sizeof(AudioSample);
    buffer.pAudioData = (byte *)pPlayerBuffer;
    buffer.pContext = pPlayerBuffer;

    HRESULT hresult = pSourceVoice->SubmitSourceBuffer(&buffer);

    if (FAILED(hresult))
        throw ref new COMException(hresult, "XAudio2SourceVoice->SubmitSourceBuffer failure");
}

AudioSample AudioFilePlayer::InterpolateSamples()
{
    double left1 = 0, left2 = 0, right1= 0, right2 = 0;

    for (int i = 0; i < 2; i++)
    {
        if (pAudioBuffer == nullptr)
            break;

        int index1 = (int)audioBufferIndex;
        int index2 = index1 + 1;
        double weight = audioBufferIndex - index1;

        if (index1 >= 0 && index1 < audioBufferLength)
        {
            left1 = (1 - weight) * pAudioBuffer[index1].Left;
            right1 = (1 - weight) * pAudioBuffer[index1].Right;
        }
        
        if (index2 >= 0 && index2 < audioBufferLength)
        {
            left2 = weight * pAudioBuffer[index2].Left;
            right2 = weight * pAudioBuffer[index2].Right;
        }
    }

    AudioSample audioSample;
    audioSample.Left = (short)(left1 + left2);
    audioSample.Right = (short)(right1 + right2);
    return audioSample;
}

// IXAudio2VoiceCallback method
void _stdcall AudioFilePlayer::OnBufferEnd(void * pContext)
{
    SubmitBuffer((AudioSample *)pContext, playerBufferLength);
}


