#include "pch.h"
#include "AudioFilePlayer.h"

using namespace StreamMusicFile;

using namespace Microsoft::WRL;
using namespace Platform;

AudioFilePlayer::AudioFilePlayer(ComPtr<IXAudio2> pXAudio2, ComPtr<IMFSourceReader> mfSourceReader)
{
    this->mfSourceReader = mfSourceReader;

    // Get the Media Foundation media type
    ComPtr<IMFMediaType> mfMediaType;
    HRESULT hresult = mfSourceReader->GetCurrentMediaType(MF_SOURCE_READER_FIRST_AUDIO_STREAM, &mfMediaType);

    if (FAILED(hresult))
        throw ref new COMException(hresult, "MFSourceReader->GetCurrentMediaType failure");

    // Create a WAVEFORMATEX from the media type
    WAVEFORMATEX* pWaveFormat;
    unsigned int waveFormatLength;
    hresult = MFCreateWaveFormatExFromMFMediaType(mfMediaType.Get(), &pWaveFormat, &waveFormatLength);

    if (FAILED(hresult))
        throw ref new COMException(hresult, "MFCreateWaveFormatExFromMFMediaType failure");

    // Create the XAudio2 source voice
    hresult = pXAudio2->CreateSourceVoice(&pSourceVoice, pWaveFormat,
                                          XAUDIO2_VOICE_NOPITCH, 1.0f, this);

    // Free the memory allocated by function
    CoTaskMemFree(pWaveFormat);

    // Submit two buffers
    SubmitBuffer();
    SubmitBuffer();

    // Start the voice playing
    pSourceVoice->Start();
    endOfFile = false;
}

void AudioFilePlayer::SubmitBuffer()
{
    // Get the next block of audio data
    int audioBufferLength;
    byte * pAudioBuffer = GetNextBlock(&audioBufferLength);

    if (pAudioBuffer != nullptr)
    {
        // Create an XAUDIO2_BUFFER for submitting audio data
        XAUDIO2_BUFFER buffer = {0};
        buffer.AudioBytes = audioBufferLength;
        buffer.pAudioData = pAudioBuffer;
        buffer.pContext = pAudioBuffer;
        HRESULT hresult = pSourceVoice->SubmitSourceBuffer(&buffer);

        if (FAILED(hresult))
            throw ref new COMException(hresult, "IXAudio2SourceVoice->SubmitSourceBuffer failure");
    }
}

byte * AudioFilePlayer::GetNextBlock(int * pAudioBufferLength)
{
    // Get an IMFSample object
    ComPtr<IMFSample> mfSample;
    DWORD flags = 0;
    HRESULT hresult = mfSourceReader->ReadSample(MF_SOURCE_READER_FIRST_AUDIO_STREAM, 
                                                 0, nullptr, &flags, nullptr, &mfSample);
    if (FAILED(hresult))
        throw ref new COMException(hresult, "MFSourceReader->ReadSample failure");

    // Check if we're at the end of the file
    if (flags & MF_SOURCE_READERF_ENDOFSTREAM)
    {
        endOfFile = true;
        *pAudioBufferLength = 0;
        return nullptr;
    }

    // If not, convert the data to a contiguous buffer
    ComPtr<IMFMediaBuffer> mfMediaBuffer;
    hresult = mfSample->ConvertToContiguousBuffer(&mfMediaBuffer);

    if (FAILED(hresult))
         throw ref new COMException(hresult, "IMFSample->ConvertToContiguousBuffer failure");

    // Lock the audio buffer and copy the samples to local memory
    uint8 * pAudioData = nullptr;
    DWORD audioDataLength = 0;
    hresult = mfMediaBuffer->Lock(&pAudioData, nullptr, &audioDataLength); 

    if (FAILED(hresult))
        throw ref new COMException(hresult, "IMFMediaBuffer->Lock failure");

    byte * pAudioBuffer = new byte[audioDataLength];
    CopyMemory(pAudioBuffer, pAudioData, audioDataLength);
    hresult = mfMediaBuffer->Unlock();

    if (FAILED(hresult))
        throw ref new COMException(hresult, "IMFMediaBuffer->Unlock failure");

    *pAudioBufferLength = audioDataLength;
    return pAudioBuffer;
}

// Callback methods from IXAudio2VoiceCallback
void _stdcall AudioFilePlayer::OnBufferEnd(void* pContext)
{
    // Remember to free the audio buffer!
    delete[] pContext;

    // Either submit a new buffer or clean up
    if (!endOfFile)
    {
        SubmitBuffer();
    }
    else
    {
        pSourceVoice->DestroyVoice();
        HRESULT hresult = MFShutdown();

        if (FAILED(hresult))
            throw ref new COMException(hresult, "MFShutdown failure");
    }
}
