﻿#include "pch.h"
#include "MainPage.xaml.h"
#include "AudioFilePlayer.h"

using namespace StreamMusicFile;

using namespace Concurrency;
using namespace Microsoft::WRL;
using namespace Platform;
using namespace Windows::Foundation;
using namespace Windows::Foundation::Collections;
using namespace Windows::Storage;
using namespace Windows::Storage::FileProperties;
using namespace Windows::Storage::Pickers;
using namespace Windows::Storage::Streams;
using namespace Windows::UI::Xaml;
using namespace Windows::UI::Xaml::Controls;
using namespace Windows::UI::Xaml::Controls::Primitives;
using namespace Windows::UI::Xaml::Data;
using namespace Windows::UI::Xaml::Input;
using namespace Windows::UI::Xaml::Media;
using namespace Windows::UI::Xaml::Media::Imaging;
using namespace Windows::UI::Xaml::Navigation;

MainPage::MainPage()
{
	InitializeComponent();

    // Create an IXAudio2 object
    HRESULT hresult = XAudio2Create(&pXAudio2);

    if (FAILED(hresult))
        ref new COMException(hresult, "XAudio2Create failure");

    // Create a mastering voice
    hresult = pXAudio2->CreateMasteringVoice(&pMasteringVoice); 

    if (FAILED(hresult))
        ref new COMException(hresult, "IXAudio2->CreateMasteringVoice failure");

    // Initialize the volume
    hresult = pMasteringVoice->SetVolume(0);

    if (FAILED(hresult))
        ref new COMException(hresult, "IXAudio2MasteringVoice->SetVolume failure");
}

void MainPage::OnVolumeSliderValueChanged(Object^ sender, RangeBaseValueChangedEventArgs^ args)
{
    HRESULT hresult = pMasteringVoice->SetVolume(float(args->NewValue / 100));

    if (FAILED(hresult))
        ref new COMException(hresult, "IXAudio2MasteringVoice->SetVolume failure");
}

void MainPage::OnOpenButtonClick(Object^ sender, RoutedEventArgs^ args)
{
    FileOpenPicker^ fileOpenPicker = ref new FileOpenPicker();
    fileOpenPicker->SuggestedStartLocation = PickerLocationId::MusicLibrary;
    fileOpenPicker->FileTypeFilter->Append(".wma");
    fileOpenPicker->FileTypeFilter->Append(".mp3");
    fileOpenPicker->FileTypeFilter->Append(".wav");
    
    // Get a music file from the picker
    task<StorageFile^> pickSingleFileTask = create_task(fileOpenPicker->PickSingleFileAsync());
    pickSingleFileTask.then([this](StorageFile^ storageFile)
    {
        if (storageFile == nullptr)
            return;

        // Get the thumbnail for the display 
        task<StorageItemThumbnail^> getThumbnailTask = create_task(storageFile->GetThumbnailAsync(ThumbnailMode::SingleItem));
        getThumbnailTask.then([this](StorageItemThumbnail^ storageItemThumbnail)
        {
            task<void> setSourceTask(bitmapImage->SetSourceAsync(storageItemThumbnail));
            setSourceTask;
        });

        // Get information about the track
        StorageItemContentProperties^ contentProperties = storageFile->Properties;
        task<MusicProperties^> getMusicPropertiesTask = create_task(contentProperties->GetMusicPropertiesAsync());
        getMusicPropertiesTask.then([this](MusicProperties^ musicProperties)
        {
            artistText->Text = musicProperties->Artist;
            albumText->Text = musicProperties->Album;
            trackText->Text = musicProperties->Title;

            int seconds = int(musicProperties->Duration.Duration / 10000000);
            wchar_t str[16];
            swprintf(str, 16, L"%i:%02i", seconds / 60, seconds % 60);
            durationText->Text = ref new String(str);
        });

        // Open the file for reading
        task<IRandomAccessStream^> openReadTask = create_task(storageFile->OpenAsync(FileAccessMode::Read));
        openReadTask.then([this](IRandomAccessStream^ randomAccessStream)
        {
            // Create an IMFSourceReader and an IXAudio2SourceVoice 
            ComPtr<IMFSourceReader> mfSourceReader = CreateSourceReader(randomAccessStream);
            new AudioFilePlayer(pXAudio2, mfSourceReader);
        });
    });
}

ComPtr<IMFSourceReader> MainPage::CreateSourceReader(IRandomAccessStream^ randomAccessStream)
{
    // Start up the Media Foundation
    HRESULT hresult = MFStartup(MF_VERSION);

    if (FAILED(hresult))
        throw ref new COMException(hresult, "MFStartup failure");

    // Create a IMFByteStream to wrap the IRandomAccessStream
    ComPtr<IMFByteStream> mfByteStream;
    hresult = MFCreateMFByteStreamOnStreamEx((IUnknown *)randomAccessStream, &mfByteStream);

    if (FAILED(hresult))
        throw ref new COMException(hresult, "MFCreateMFByteStreamOnStreamEx failure");

    // Create an attribute for low latency operation
    ComPtr<IMFAttributes> mfAttributes;
    hresult = MFCreateAttributes(&mfAttributes, 1);

    if (FAILED(hresult))
        throw ref new COMException(hresult, "MFCreateAttributes failure");

    hresult = mfAttributes->SetUINT32(MF_LOW_LATENCY, TRUE);

    if (FAILED(hresult))
        throw ref new COMException(hresult, "IMFAttributes->SetUINT32 failure");

    // Create the IMFSourceReader
    ComPtr<IMFSourceReader> mfSourceReader;
    hresult = MFCreateSourceReaderFromByteStream(mfByteStream.Get(), mfAttributes.Get(), &mfSourceReader);

    if (FAILED(hresult))
        throw ref new COMException(hresult, "MFCreateSourceReaderFromByteStream failure");

    // Create an IMFMediaType for setting the desired format
    ComPtr<IMFMediaType> mfMediaType;
    hresult = MFCreateMediaType(&mfMediaType);

    if (FAILED(hresult))
        throw ref new COMException(hresult, "MFCreateMediaType failure");

    hresult = mfMediaType->SetGUID(MF_MT_MAJOR_TYPE, MFMediaType_Audio);

    if (FAILED(hresult))
        throw ref new COMException(hresult, "MFMediaType->SetGUID failure");

    hresult = mfMediaType->SetGUID(MF_MT_SUBTYPE, MFAudioFormat_Float);

    if (FAILED(hresult))
        throw ref new COMException(hresult, "MFMediaType->SetGUID failure");

    // Set the media type in the source reader
    hresult = mfSourceReader->SetCurrentMediaType(MF_SOURCE_READER_FIRST_AUDIO_STREAM, 
                                                  0, mfMediaType.Get());
    if (FAILED(hresult))
        throw ref new COMException(hresult, "MFSourceReader->SetCurrentMediaType failure");

    return mfSourceReader;
}

